import os
import sys
import sqlite3
from sqlite3 import Error
import math
import binascii
import json
import time

# WORK IN PROGRESS

"""
The data to download is loaded in chunks, so first of all we'll need to know the size of the data
This can vary over time, so we'll need to checkwhich version each node has, and restart if it's newer.
Itchy problem: how to prevent version poisoning? E.g an evil node saying it has version 999 and feeds us garbage...

For now:
Split up the data to load in N chunks and populate a small database with one entry each:
chunks: id, start_byte, size (bytes to download), state [IDLE, TRYING, DONE]

We'll use the spawned threads to try to download any "IDLE", then any "TRYING" until all are "DONE"

"""



class Download():

    # Downloads a file from a link, uses multiple threads, combines it all in the end
    # if reload is set to True, then we will perform a clean new download
    def __init__(self, link_file, dest_folder, reload=False):
        link_file = link_file.replace('\\', '/')
        dest_folder = dest_folder.replace('\\', '/')

        # Get the data out of the link file
        with open(link_file) as json_data:
            print('get json')
            d = json.load(json_data)
            # todo: save off addresses for later use in swicharoo etc:
            # if 'address_list' in d:
            #    save_off_addresses(d['address_list'])

            # Start the actual downloading
            if not 'address_list' in d or not 'key' in d or not 'iv' in d or not 'data_rsa_pub' in d:
                print("#Error: Missing information in link file, aborting...")
                os.exit(-1)

            self.address_list = d['address_list']

            # this is the key,iv for the encrypted file we want to retrieve
            self.key = bytes.fromhex(d['key'])
            self.iv = bytes.fromhex(d['iv'])
            self.data_rsa_pub = d['data_rsa_pub']

        # Setup
        self.temp_folder = link_file + '_download/'
        if not os.path.exists(self.temp_folder):
            os.makedirs(self.temp_folder)
        self.link_file = link_file
        self.database_name = './' + self.temp_folder + link_file + '_wip.db'

        if reload and os.path.isfile(self.database_name):  # Wipe the database
            os.remove(self.database_name)

        # generate, or regenerate database if needed
        if not os.path.isfile(self.database_name):
            connection = self.get_connection()
            connection.execute('''CREATE TABLE chunks
                                  (id INTEGER PRIMARY KEY AUTOINCREMENT,
                                   start_byte              INTEGER,
                                   size                    INTEGER,
                                   state                   TEXT NULL DEFAULT "IDLE"
                                  );''')

            # Here we share data between threads so one entry only. -1 means it is not initialized yet
            connection.execute('''CREATE TABLE data
                                  (file_size               INTEGER DEFAULT -1,
                                   file_version            INTEGER DEFAULT -1
                                  );''')

            cursor = connection.cursor()
            cursor.execute('INSERT INTO data (file_size, file_version) VALUES(?, ?)', (-1, -1,))
            connection.commit()

        self.insert_chunk(0, 1024)
        self.insert_chunk(1024, 76)

        # Fork off:
        forks = 10
        for i in range(0, forks):
            new_pid = os.fork()
            if new_pid == 0:
                self.fork_main(i)

        time.sleep(8)


    def fork_main(self, nr):
        print("A fork:", nr)
        time.sleep(0.5)

    def query_one(self, query, extra_data=None):
        connection = self.get_connection()
        cursor = connection.cursor()
        if extra_data is None:
            cursor.execute(query)
        else:
            cursor.execute(query, extra_data)
        row = cursor.fetchone()
        if row is None:
            return None

        if len(row) == 1:
            return row[0]
        return None


    def insert_chunk(self, start_byte, size):
        connection = self.get_connection()
        cursor = connection.cursor()
        cursor.execute('INSERT INTO chunks (start_byte, size) VALUES(?, ?)', (start_byte, size, ))
        connection.commit()

    def get_random_chunk(self):
        # Get an IDLE one first, or a TRYING one if there are no more IDLE. Returns None if no left
        result = self.query_one('SELECT id FROM chunks WHERE state="IDLE" ORDER BY RANDOM() LIMIT 1')
        if result is not None:
            return result
        return self.query_one('SELECT id FROM chunks WHERE state="TRYING" ORDER BY RANDOM() LIMIT 1')

    def get_connection(self):
        try:
            connection = sqlite3.connect(self.database_name)
            return connection
        except Error:
            print(Error)
        sys.exit(-3)

    # returns the number of chunks left to download or zero if we are done downloading
    def download_chunks_left(self, database_name):
        # Open database

        # Check how many chunks that are not finished
        return 1

    def download_chunk(self, database_name, name, data_rsa_pub, chunk_size):
        # Get a random chunk to download

        # Grab a random node

        # Try to download chunk

        # During download, check if DB either says we're done, or if it doesn't exist anymore, if so just close socket and bail out, we're done

        # If success, save it off and tick the DB entry as 'DONE', else just close socket and return
        pass


    def get_data_parallel(self, address_list, name, data_rsa_pub):  # Note: You do the decrypting!
        # todo: if the version wee get from a node is > our versionm then redo everything with this new version!

        # Get the version + filesize

        # Make a database with entries for N chunks
        chunk_size = 1024 * 1024  # todo: make it configurable
        max_chunks = 100  # this might override chunksize

        # todo: do this
        version = 1
        file_size = 123456

        if chunk_size * max_chunks < file_size:
            chunk_size = int(math.ceil(file_size / max_chunks))

        chunks = int(math.ceil(file_size/chunk_size))
        print("Preparing " + str(chunks) + " chunks for downloading")


        # DB: id, chunkstatus = { 'NOT_STARTED', 'DOWNLOADING', 'DONE'  }


        # For now just do it single threaded until it works:

        while(self.download_chunks_left(self.database_name) > 0):
            self.download_chunk(name, data_rsa_pub, chunk_size)


download = Download('./linuxmusic1.10f', './', True)